<?php
/**
 * The PhoneKey API class to communicate with the PhoneKey Webservice
 * @version 0.1
 *
 */
class PhoneKeyAPI{

	/*
	 * Values loaded from the config file
	 */
	private $api_key;
	private $private_key;
	private $remote_host_address;
	private $remote_host_port;
	private $remote_host_public_key;

	/**
	 * A path to a config file must be given to construct the API
	 * @param unknown $userConfigFile
	 */
	public function __construct($userConfigFile){
		// We try to open the config file
		if(($f = @fopen($userConfigFile, "r")) !== false){
			// We load & decode the content of the config file
			$array = json_decode(fread($f, filesize($userConfigFile)), true);
			if($array !== null){
				// We check tha all the config is ok, we must have 4 variables
				if(isset($array["api_key"]) && isset($array["private_key"]) && isset($array["remote_host_address"]) && isset($array["remote_host_port"]) && isset($array["remote_host_public_key"])){
					$this->api_key = $array["api_key"];
					$this->remote_host_address = $array["remote_host_address"];
					$this->remote_host_port = $array["remote_host_port"];

					// Loading our private key and the WebService public key
					$this->private_key = openssl_get_privatekey($array["private_key"]);
					$this->remote_host_public_key = openssl_get_publickey($array["remote_host_public_key"]);
				}
				// If we are here almost all the step of the config checking & loading are validated
			}
		}
	}

	/**
	 * Check the runtime and the config file of the API
	 * @return multitype:string : an array containing all the errors found, count($returned_array) == 0 if no error found
	 */
	public static function checkEnvironment($userConfigFile){
		// We check first all the required PHP extensions
		$errors = array();
		$extensions = get_loaded_extensions(false);
		if(!extension_loaded("json"))
			$errors[] = "Json extension not loaded";
		if(!extension_loaded("curl"))
			$errors[] = "Curl extension not loaded";
		if(!extension_loaded("openssl"))
			$errors[] = "OpenSSl extension not loaded";

		// We try to open the config file
		if(($f = @fopen($userConfigFile, "r")) === false)
			$errors[] = "Unable to load '".$userConfigFile."'. File missing ?";
		else{
			// We load & decode the content of the config file
			$array = json_decode(fread($f, filesize($userConfigFile)), true);
			if($array === null)
				$errors[] = "Unable to parse the config file. Is it json ?";
			else{
				// We check the api key
				if(!isset($array["api_key"])){
					$errors[] = "The key 'api_key' in the config file is missing.";
				}else if(!preg_match("#^[a-z0-9]{30}$#", $array["api_key"])){
					$errors[] = "The value of the 'api_key' key in the config file is not valid.";
				}

				// We check the private key
				if(!isset($array["private_key"])){
					$errors[] = "The key 'private_key' in the config file is missing.";
				}else if(@openssl_get_privatekey($array["private_key"]) === false){
					$errors[] = "Unable to load the private key. Is the format OK ?";
				}

				// We check the webservice public key
				if(!isset($array["remote_host_public_key"])){
					$errors[] = "The key 'remote_host_public_key' in the config file is missing.";
				}
				else if(@openssl_get_publickey($array["remote_host_public_key"]) === false){
					$errors[] = "Unable to load the public key. Is the format OK ?";
				}

				if(!isset($array["remote_host_address"])){
					$errors[] = "The key 'remote_host_address' in the config file is missing.";
				}
				else if(!preg_match("#^[a-zA-Z0-9\.]+[a-zA-Z]+$#", $array["remote_host_address"])){
					$errors[] = "Bad format of the remote host address. Must be : (subsubdomain.)subdomain.domain";
				}

				// We check the remote host port
				if(!isset($array["remote_host_port"])){
					$errors[] = "The key 'remote_host_port' in the config file is missing.";
				}else if(!preg_match("#^[0-9]+$#", $array["remote_host_port"]) || $array["remote_host_port"] < 1 || $array["remote_host_port"] > 65535){
					$errors[] = "The value of the 'remote_host_port' key in the config file is not valid (1-65535).";
				}
			}
		}
		return $errors;
	}

	/**
	 * Check the WebService without throwing any PhoneKeyServiceException
	 * @return boolean|string true if all is ok, otherwise a string detailing the cause if the error
	 */
	public function checkService(){
		try{
			$msg = $this->decryptData($this->processRequest(array("request" => "I want to know if all is ok between us :p"), "checkservice"), $this->remote_host_public_key, $this->private_key);
			if(isset($msg["response"]) && $msg["response"] === "All is ok, I can hear you my dear :)")
				return true;
			else
				return "Unexpected response string (".$msg.")";
		}catch(PhoneKeyServiceException $e){
			return $e->getMessage();
		}
	}

	/**
	 * Ask a temporary token for an authentication
	 * Phone number format : (+33|0)000000000
	 * Each digit can be separated from the other with '.', ' ' or '-' 
	 * @param string $phoneNumber
	 * @return array the token
	 */
	public function requestLocalAuthentication($phoneNumber){
		// Processing request
		$msg = $this->decryptData($this->processRequest(array("phoneNumber" => $phoneNumber), "newlocaltoken"), $this->remote_host_public_key, $this->private_key);
		// Check of the response integrity
		$this->checkArrayKeys($msg, array("token","type","creationDate","duration","timeleft","validated","clientPhoneNumber","phoneKeyPhoneNumber","phoneCode","messagesCount","onSuccessURL","onAbortURL","onTimeOutURL"));
		return $msg;
	}

	/**
	 * Ask for a temporary token for a remote authentication. PhoneKey will ask his phone number to the client, so you don't have to provide it.
	 * The phone number of the client will not be able in the token status informations.
	 * You have to provide 3 URL on which webservice.phonekey.eu will redirect the client if the token has (been validated| timed out| been aborted)
	 * URLs must begin with http:// or https://
	 * Note that you don't have to provide the phonenumber of your client, PhoneKey.eu will do it 
	 * @param string $onSuccessURL an url on which redirect the client if he was successfully authenticated
	 * @param string $onAbortURL an url on which redirect the client if he aborted the authentication
	 * @param string $onTimeOutURL an url on which redirect the client if the authentication timed out
	 * @return array the token
	 */
	public function requestRemoteAuthentication($onSuccessURL, $onAbortURL, $onTimeOutURL){
		// Processing request
		$msg = $this->decryptData($this->processRequest(array("onSuccessURL" => $onSuccessURL, "onAbortURL" => $onAbortURL, "onTimeOutURL" => $onTimeOutURL), "newremotetoken"), $this->remote_host_public_key, $this->private_key);
		// Check of the response integrity
		$this->checkArrayKeys($msg, array("token","type","creationDate","duration","timeleft","validated","clientPhoneNumber","phoneKeyPhoneNumber","phoneCode","messagesCount","onSuccessURL","onAbortURL","onTimeOutURL"));
		return $msg;
	}
	
	/**
	 * Ask for a temporary token for a remote authentication. You have to provide the phone number of the client you want that PhoneKey authenticate for you.
	 * The phone number of the client will not be able in the token status informations.
	 * You have to provide 3 URL on which webservice.phonekey.eu will redirect the client if the token has (been validated| timed out| been aborted)
	 * URLs must begin with http:// or https://
	 * Phone number format : (+33|0)000000000
	 * Each digit can be separated from the other with '.', ' ' or '-' 
	 * @param $phoneNumber the phone number of the client
	 * @param string $onSuccessURL an url on which redirect the client if he was successfully authenticated
	 * @param string $onAbortURL an url on which redirect the client if he aborted the authentication
	 * @param string $onTimeOutURL an url on which redirect the client if the authentication timed out
	 * @return array the token
	 */
	public function requestRemoteAuthenticationWithPhoneNumber($phoneNumber, $onSuccessURL, $onAbortURL, $onTimeOutURL){
		// Processing request
		$msg = $this->decryptData($this->processRequest(array("phoneNumber" => $phoneNumber, "onSuccessURL" => $onSuccessURL, "onAbortURL" => $onAbortURL, "onTimeOutURL" => $onTimeOutURL), "newremotetokenwithnumber"), $this->remote_host_public_key, $this->private_key);
		// Check of the response integrity
		$this->checkArrayKeys($msg, array("token","type","creationDate","duration","timeleft","validated","clientPhoneNumber","phoneKeyPhoneNumber","phoneCode","messagesCount","onSuccessURL","onAbortURL","onTimeOutURL"));
		return $msg;
	}

	/**
	 * Get the status of a given token
	 * @param string $tokenCode
	 * @throws PhoneKeyServiceException if something goes wrong
	 * @return array:string informations about the token
	 */
	public function getStatus($tokenCode){
		// Processing request
		$msg = $this->decryptData($this->processRequest(array("token" => $tokenCode), "tokenstatus"), $this->remote_host_public_key, $this->private_key);
		// Check of the response integrity
		$this->checkArrayKeys($msg, array("token","type","creationDate","duration","timeleft","validated","clientPhoneNumber","phoneKeyPhoneNumber","phoneCode","messagesCount","onSuccessURL","onAbortURL","onTimeOutURL", "remoteURL"));
		return $msg;
	}

	/**
	 * It sends a POST request to the WebService and return the response.
	 * The $content is transformed into Json, is crypted and then sent to the WebService.
	 * Then, the reponse is decrypted, checked and returned.
	 * This function handle the security and integrity layer of the communication with the WebService.
	 * @param array $data the content of the request : an associative array
	 * @param $uriSuffix the suffix of the URI
	 * @throws PhoneKeyServiceException if a problem between the client and the server is detected
	 */
	private function processRequest(array $data, $uriSuffix){
		// Crypting the data
		$req = $this->encryptData($data, $this->remote_host_public_key, $this->private_key);

		// Preparation of the request
		$ch = curl_init("http://".$this->remote_host_address."/webservice/".$this->api_key."/".$uriSuffix);
		curl_setopt($ch, CURLOPT_PORT, $this->remote_host_port);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
		curl_setopt($ch, CURLOPT_USERAGENT, "Phonekey API v0.1a");
		curl_setopt($ch, CURLOPT_POST, true);
		curl_setopt($ch, CURLOPT_POSTFIELDS, array("data" => $req));

		// Execution of the request
		$response = curl_exec($ch);
		// Check of the http response code
		if(curl_getinfo($ch, CURLINFO_HTTP_CODE) != 200){
			throw new PhoneKeyServiceException(PhoneKeyServiceException::EX_WEBSERVICE_SPECIFIC_ERROR, "[http code ".curl_getinfo($ch, CURLINFO_HTTP_CODE)."] / response : ".$response);
		}

		// If all is ok, we decode the content of the response and return it as an array
		return $response;
	}

	/**
	 * It crypts the data with the public and private keys provided
	 * @param unknown $data the data to be encrypted, must be an array
	 * @param unknown $publicKey the public key used to crypt the key of the first symetric encryption
	 * @param unknown $privateKey the private key used to crypt the key of the second symetric encryption
	 * @throws PhoneKeyServiceException if something goes wrong
	 * @return string the encrypted data
	 */
	private function encryptData($data, $publicKey, $privateKey){
		// Deflating data1 into Json
		$this->rwseof($data1 = json_encode($data), PhoneKeyServiceException::EX_JSON_DEFLATIN_ERROR);

		// We generate a "symetric" key n°1 (sk stands for symetric key)
		$strong = true;
		$d = date("YmdHis");
		$key1 = $this->rwseof(openssl_random_pseudo_bytes(96, $strong), PhoneKeyServiceException::EX_RANDOM_BYTES_FAILED);
		$key1 = $d.$key1;

		// We generate a "symetric" key n°2
		$strong = true;
		$key2 = $this->rwseof(openssl_random_pseudo_bytes(96, $strong), PhoneKeyServiceException::EX_RANDOM_BYTES_FAILED);
		$key2 = $d.$key2;

		// We encrypt the "symetric" key n°1 with our private key
		$this->rwseof(openssl_private_encrypt($key1, $skey1, $privateKey), PhoneKeyServiceException::EX_BAD_KEY1_ENCRYPTING);

		// We encode the encrypted sk1 in base64
		$skey1_b64 = $this->rwseof(base64_encode($skey1), PhoneKeyServiceException::EX_BAD_SKEY1_BASE64_ENCODING);
		if(($n = substr_count($skey1_b64, "=")) < 2){
			$skey1_b64 .= $n == 1 ? "=" : "==";
		}

		// We encrypt the "symetric" key 2 with the webservice public key
		$this->rwseof(openssl_public_encrypt($key2, $skey2, $publicKey), PhoneKeyServiceException::EX_BAD_KEY2_ENCRYPTING);
		// We encode encrypted sk2 in base 64
		$skey2_b64 = $this->rwseof(base64_encode($skey2), PhoneKeyServiceException::EX_BAD_SKEY2_BASE64_ENCODING);
		if(($n = substr_count($skey2_b64, "=")) < 2){
			$skey2_b64 .= $n == 1 ? "=" : "==";
		}

		// We crypt the data1 with sk1
		$encrypted_data1 = $this->rwseof(openssl_encrypt($data1, "aes-256-cbc", $key1, OPENSSL_RAW_DATA, substr($key1, 0, 16)), PhoneKeyServiceException::EX_BAD_DATA1_ENCRYPTING);

		// We encode the encrypted data1 in base64
		$sdata1_b64 = $this->rwseof(base64_encode($encrypted_data1), PhoneKeyServiceException::EX_BAD_SDATA1_BASE64_ENCODING);
		if(($n = substr_count($sdata1_b64, "=")) < 2){
			$sdata1_b64 .= $n == 1 ? "=" : "==";
		}

		// We concatenate the encrypted base 64 data1 and the encrypted base 64 sk1
		$data2 = $sdata1_b64.$skey1_b64;

		// We crypt data2 with sk2
		$sdata2 = $this->rwseof(openssl_encrypt($data2, "aes-256-cbc", $key2, OPENSSL_RAW_DATA, substr($key2, 0, 16)), PhoneKeyServiceException::EX_BAD_DATA2_CRYPTING);

		// We encode encrypted data 2 in base 64
		$sdata2_b64 = $this->rwseof(base64_encode($sdata2), PhoneKeyServiceException::EX_BAD_SDATA2_BASE64_ENCODING);
		if(($n = substr_count($sdata2_b64, "=")) < 2){
			$sdata2_b64 .= $n == 1 ? "=" : "==";
		}

		return $sdata2_b64.$skey2_b64;
	}

	/**
	 * It decrypts a request from a client, rising exception if something is going wrong
	 * @param unknown $data
	 * @param unknown $publicKey
	 * @param unknown $privateKey
	 * @throws PhoneKeyServiceException
	 * @return unknown
	 */
	private function decryptData($data, $publicKey, $privateKey){
		// We separate the sdata2 (base64) and the spk2 (base 64) from $data
		$first = strpos($data, "=");
		$last = strlen($data);
		if($first === false || $data[$first+1] !== "=" || $data[$last-1] !== "=" || $data[$last-2] !== "="){
			throw new PhoneKeyServiceException(PhoneKeyServiceException::EX_BAD_DATA2_CHUNKING, $data);
		}
		$sdata2_b64 = substr($data, 0, $first+2);
		$skey2_b64 = substr($data, $first+2, $last);

		// base 64 decoding the sdata2
		$sdata2 = $this->rwseof(base64_decode($sdata2_b64), PhoneKeyServiceException::EX_MALFORMED_REQUEST);

		// base 64 decoding the skey2
		$skey2 = $this->rwseof(base64_decode($skey2_b64), PhoneKeyServiceException::EX_MALFORMED_REQUEST);

		// Decrypting skey2
		$this->rwseof(openssl_private_decrypt($skey2, $key2, $this->private_key), PhoneKeyServiceException::EX_BAD_SKEY2_DECRYPTING);

		// Decrypting sdata2
		$data2 = $this->rwseof(openssl_decrypt($sdata2, "aes-256-cbc", $key2, OPENSSL_RAW_DATA, substr($key2, 0, 16)), PhoneKeyServiceException::EX_BAD_SDATA2_DECRYPTING);

		// We separate the data1 and the pk1 from the decrypted and base64 decoded request
		$first = strpos($data2, "=");
		$last = strlen($data2);
		if($first === false || $data2[$first+1] !== "=" || $data2[$last-1] !== "=" || $data2[$last-2] !== "="){
			throw new PhoneKeyServiceException(PhoneKeyServiceException::EX_BAD_DATA1_CHUNKING);
		}
		$sdata1_b64 = substr($data2, 0, $first+2);
		$skey1_b64 = substr($data2, $first+2, $last);

		// base 64 decoding of sdata1
		$sdata1 = $this->rwseof(base64_decode($sdata1_b64), PhoneKeyServiceException::EX_BAD_SDATA1_BASE64_DECODING);

		// base 64 decoding of skey1
		$skey1 = $this->rwseof(base64_decode($skey1_b64), PhoneKeyServiceException::EX_BAD_SKEY1_BASE64_DECODING);

		// Decrypting skey1
		$this->rwseof(openssl_public_decrypt($skey1, $key1, $publicKey), PhoneKeyServiceException::EX_BAD_SKEY1_DECRYPTING);

		// Decrypting sdata1
		$data1 = $this->rwseof(openssl_decrypt($sdata1, "aes-256-cbc", $key1, OPENSSL_RAW_DATA, substr($key1, 0, 16)), PhoneKeyServiceException::EX_BAD_SDATA1_DECRYPTING);

		// Inflating from json
		return $this->rwseof(json_decode($data1, true), PhoneKeyServiceException::EX_BAD_DATA1_INFLATING);
	}

	/**
	 * rwseof stands for : Raise PhoneKeyServiceExeption On False.
	 * Throws a PhoneKeyServiceExeption if $stuff === false, otherwise return $stuff
	 * @param unknown $stuff
	 * @param unknown $errorCode the code of the WebServiceException
	 * @throws PhoneKeyServiceException
	 */
	private function rwseof($stuff, $errorCode){
		if($stuff === false)
			throw new PhoneKeyServiceException($errorCode);
		else
			return $stuff;
	}

	/**
	 * Throws an exception if a key from $keys doesn't exist in the $array
	 * @param unknown $array
	 * @param unknown $keys
	 * @throws PhoneKeyServiceException
	 */
	private function checkArrayKeys(array $array, array $keys){
		foreach($keys as $key){
			if(!array_key_exists($key, $array))
				throw new PhoneKeyServiceException(PhoneKeyServiceException::EX_WEBSERVICE_SPECIFIC_ERROR, "Missing response value(s) (at least '".$key."')");
		}
	}
}